#include "RendererVisLabGlyphExercise.h"
#include "vrc_plugin_renderer_vislabglyphexercise.cpp"
#include "svn_plugin_renderer_vislabglyphexercise.cpp"
#include "CoreVersion.h"
#include "Volume.h"

#include "ResourceTexture3D.h"

DEFINE_RENDERERPLUGIN(RendererVisLabGlyphExercise)
DEFINE_PLUGIN_VERSION(RendererVisLabGlyphExercise,0,0,1)

/**
 * @class RendererVisLabGlyphExercise
 * @brief Renderer plugin.
 *
 * [TODO: class description]
 */

/**
 * @brief Constructor
 */
RendererVisLabGlyphExercise::RendererVisLabGlyphExercise(Plugin & pluPlugin) : PluginInstance(pluPlugin), m_uVertexBuffer(0),m_bUpdateVertexBuffer(true)
{
	if (!GLEW_ARB_fragment_shader)
		throw Exception("GL_ARB_fragment_shader is not supported by your hardware or driver.");

	if (!GLEW_ARB_vertex_shader)
		throw Exception("GL_ARB_vertex_shader is not supported by your hardware or driver.");

	PLUGIN_LOADSHADER(m_shaGlyphs,"plugin_renderer_vislabglyphexercise.glsl");

	m_modObserverUpdate.connect(this,&RendererVisLabGlyphExercise::modifiedUpdate);
	m_modObserverVolume.connect(this,&RendererVisLabGlyphExercise::modifiedVolume);

	m_shaGlyphs.addObserver(&m_modObserverUpdate);

	GetPlugin().GetProperty("Projection Transformation").require(Variant::TypeMatrix()).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Projection Transformation","A projection transformation matrix.");

	GetPlugin().GetProperty("Viewing Transformation").require(Variant::TypeMatrix()).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Viewing Transformation","A viewing transformation matrix.");

	GetPlugin().GetProperty("Transformation").require(Variant::TypeMatrix()).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Transformation","Specifies the volume's transformation");

	GetPlugin().GetProperty("Volume").require(Variant::TypeHandle()).addObserver(&m_modObserverVolume);
	GetPlugin().SetPropertyDescription("Volume","A volume resource.");

	GetPlugin().GetProperty("Bounding Box").require(Variant::TypeBox()).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Bounding Box","Specifies the volume's bounding box if 'Use Bounding Box' is enabled.");

	GetPlugin().GetProperty("Use Bounding Box").require(Variant::TypeBoolean(false)).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Use Bounding Box","If enabled, the property 'Bounding Box' will be used as the volume's bounding box.");

	GetPlugin().GetProperty("Glyph Density").require(Variant::TypeFloat(0.21f,0.01f,1.0f)).addObserver(&m_modObserverVolume);
	GetPlugin().SetPropertyDescription("Glyph Density","Controls how many glpyhs will be mapped to the volume.");

	GetPlugin().GetProperty("Glyph Threshold").require(Variant::TypeFloat(0.37f,0.0f,1.0f)).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Glyph Threshold","The minimum value the volume has to assume for a glyph to be displayed");

	GetPlugin().GetProperty("Gradient Elongation Scale").require(Variant::TypeFloat(8.0f,1.0f,15.0f)).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Gradient Elongation Scale","A Scale factor for the gradient visualisation, controls the elongation amount");

	GetPlugin().GetProperty("Light Direction").require(Vector(0.0,0.0,1.0f).GetNormalized()).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Light Direction","Specifies the light direction for shading.");

	GetPlugin().GetProperty("Shading Intensity").require(Variant(1.0f,0.0f,1.0f)).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Shading Intensity","Controls the overall shading intensity.");

	GetPlugin().GetProperty("Diffuse Shading").require(Variant(0.6f,0.0f,1.0f)).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Diffuse Shading","Controls the amount of diffuse shading.");

	GetPlugin().GetProperty("Specular Shading").require(Variant(0.55f,0.0f,1.0f)).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Specular Shading","Controls the amount of specular shading.");

	GetPlugin().GetProperty("Specular Exponent").require(Variant(32.0f,0.0f,128.0f)).addObserver(&m_modObserverUpdate);
	GetPlugin().SetPropertyDescription("Specular Exponent","Controls the specular shading exponent, i.e., the shininess of the material.");

	m_fInteractionTimeout = 0.33f;
	m_fLastRenderTime = 0.0f;
}

/**
 * @brief Destructor
 */
RendererVisLabGlyphExercise::~RendererVisLabGlyphExercise()
{
	if (m_uVertexBuffer)
		glDeleteBuffers(1, &m_uVertexBuffer );
}

/**
 * @copydoc Renderer::idle()
 */
void RendererVisLabGlyphExercise::idle()
{
	const Handle hanVolume = GetPlugin().GetProperty("Volume");
	VolumeResource *pVolume = hanVolume.GetResource<VolumeResource>();

	//if volume not loaded
	if (!pVolume || pVolume->GetSize().GetMaximum() == 0)
		return;

	if (m_fDesiredRenderTime > 0.0f)
	{
		if (m_timInteractionTimer.IsRunning())
		{
			if (float(m_timInteractionTimer) >= m_fInteractionTimeout)
			{
				GetPlugin().update();
			}
		}
	}
}

/**
 * @copydoc Renderer::display()
 */
void RendererVisLabGlyphExercise::display(Canvas & canCanvas)
{
	if (!m_uVertexBuffer)
		glGenBuffers( 1, &m_uVertexBuffer );

	if(m_bUpdateVertexBuffer)
		updateVertexBuffer();

	glFinish();

	Timer timTimer;
	timTimer.start();

	const Handle hanVolume = GetPlugin().GetProperty("Volume");
	VolumeResource *pVolume = hanVolume.GetResource<VolumeResource>();

	if (!pVolume || pVolume->GetSize().GetMaximum() == 0)
		return;

	m_shaGlyphs.bind();

	// Default transformation
	Matrix matTransformation = Matrix();
	matTransformation = pVolume->GetProperty("Transformation",matTransformation);

	canCanvas.bind();

	Matrix matProjectionTransformation = GetPlugin().GetProperty("Projection Transformation");
	const Matrix matViewingTransformation = GetPlugin().GetProperty("Viewing Transformation");
	
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(matProjectionTransformation.Get());

	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(matViewingTransformation.Get());

	//For making the glyph billboards large enough in the shader, we need to calculate the extent of a glyph bounding box in viewspace
	//Later in the shader, they will still be enlarged if necessary due to gradient elongation
	Vector vBboxExtent = Vector(m_fPointSize*0.5f, m_fPointSize*0.5f, m_fPointSize*0.5f);
	Box dummyBbox = Box(-vBboxExtent,  vBboxExtent);
	Vector vecViewGlyphQuadMin = dummyBbox.GetMinimum(matViewingTransformation);
	Vector vecViewGlyphQuadMax = dummyBbox.GetMaximum(matViewingTransformation);
	float fExtentInViewspace = (vecViewGlyphQuadMax-vecViewGlyphQuadMin).GetMaximum();

	// Default bounding box of volume
	Box boxBounds = pVolume->GetBounds().GetTranslated(-pVolume->GetBounds().GetCenter());
	boxBounds = pVolume->GetProperty("Bounding Box",boxBounds);

	const bool bUseBoundingBox = GetPlugin().GetProperty("Use Bounding Box");

	if (bUseBoundingBox)
		boxBounds = pVolume->GetProperty("Bounding Box",boxBounds);


	Matrix matNormalRotationScale = matTransformation;
	matNormalRotationScale.scale(boxBounds.GetExtent() / pVolume->GetSize());

	const Matrix matNormalTransformation = (matViewingTransformation * matNormalRotationScale.GetRotationMatrix() * matNormalRotationScale.GetScaleMatrix().GetInverse());

	// Bind volume texture
	glActiveTexture(GL_TEXTURE0);	
	ResourceTexture::Get<VolumeTexture3D>(pVolume).bind();
	glUniform1i(m_shaGlyphs.GetUniformLocation("samVolume"), 0);

	Vector vecCamPos = matViewingTransformation.GetInverse().GetTranslation();

	glUniform1f(m_shaGlyphs.GetUniformLocation("fGlyphThreshold"), m_fGlyphThreshold);
	glUniform3f(m_shaGlyphs.GetUniformLocation("vecCameraPosition"), vecCamPos[0], vecCamPos[1], vecCamPos[2]);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fExtentInWorldspace"), m_fPointSize);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fExtentInViewspace"), fExtentInViewspace);

	glUniform1f(m_shaGlyphs.GetUniformLocation("fGradientScale"), m_fGradientScale);

	glUniformMatrix4fv(m_shaGlyphs.GetUniformLocation("matProjection"),1,false,matProjectionTransformation.Get());
	glUniformMatrix4fv(m_shaGlyphs.GetUniformLocation("matViewing"),1,false,matViewingTransformation.Get());
	glUniformMatrix4fv(m_shaGlyphs.GetUniformLocation("matInverseViewing"),1,false,matViewingTransformation.GetInverse().Get());

	glUniformMatrix4fv(m_shaGlyphs.GetUniformLocation("matNormalInverse"),1,false,matNormalTransformation.GetInverse().Get()); //get matrix that transforms normals from view to world space

	const Vector vecLightDirection = GetPlugin().GetProperty("Light Direction");
	const float fShadingIntensity = GetPlugin().GetProperty("Shading Intensity");
	const float fDiffuseShading = GetPlugin().GetProperty("Diffuse Shading");
	const float fSpecularShading = GetPlugin().GetProperty("Specular Shading");
	const float fSpecularExponent = GetPlugin().GetProperty("Specular Exponent");
	glUniform3f(m_shaGlyphs.GetUniformLocation("vecLight"), vecLightDirection[0], vecLightDirection[1], vecLightDirection[2]);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fShadingIntensity"), fShadingIntensity);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fDiffuseShading"), fDiffuseShading);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fSpecularShading"), fSpecularShading);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fSpecularExponent"), fSpecularExponent);

	glUniform1f(m_shaGlyphs.GetUniformLocation("fStartX"), m_fStartX);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fStartY"), m_fStartY);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fStartZ"), m_fStartZ);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fRangeX"), m_fRangeX);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fRangeY"), m_fRangeY);
	glUniform1f(m_shaGlyphs.GetUniformLocation("fRangeZ"), m_fRangeZ);

	glBindBuffer(GL_ARRAY_BUFFER, m_uVertexBuffer);
	glVertexPointer( 3, GL_FLOAT, 0, NULL);
	glEnableClientState(GL_VERTEX_ARRAY);	

	glDrawArrays(GL_POINTS, 0, m_uElementCount);

	glDisableClientState( GL_VERTEX_ARRAY );	
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	canCanvas.release();
	
	m_shaGlyphs.release();
	ResourceTexture::Get<VolumeTexture3D>(pVolume).release();

	glFinish();
	m_fLastRenderTime = float(timTimer);	
}

/**
 * @copydoc Renderer::overlay()
 */
void RendererVisLabGlyphExercise::overlay(Canvas & canCanvas)
{

}

/**
 * @copydoc Renderer::reshape()
 */
void RendererVisLabGlyphExercise::reshape(const unsigned int uWidth, const unsigned int uHeight)
{
	if (uWidth != m_uWidth || uHeight != m_uHeight)
	{
		m_uWidth = uWidth;
		m_uHeight = uHeight;
		GetPlugin().update();
	}
}

/**
 * @brief Event handler which triggers a re-render.
 */
void RendererVisLabGlyphExercise::modifiedUpdate( const Observable & obsObservable, const Observable::Event & eveEvent )
{
	if (m_fDesiredRenderTime > 0.0f)
		m_timInteractionTimer.start();

	m_fGlyphThreshold = GetPlugin().GetProperty("Glyph Threshold");
	m_fGradientScale = GetPlugin().GetProperty("Gradient Elongation Scale");


	GetPlugin().update();
}


/**
 * @brief Event handler which triggers a rebuild of the glyph set.
 */
void RendererVisLabGlyphExercise::modifiedVolume( const Observable & obsObservable, const Observable::Event & eveEvent )
{
	if (m_fDesiredRenderTime > 0.0f)
		m_timInteractionTimer.start();
	
	m_bUpdateVertexBuffer = true;

	GetPlugin().update();
}

/**
 * @brief Perform update of the glyph set.
 */
void RendererVisLabGlyphExercise::updateVertexBuffer()
{
	if(m_bUpdateVertexBuffer)
	{
		const Handle hanVolume = GetPlugin().GetProperty("Volume");
		VolumeResource *pVolume = hanVolume.GetResource<VolumeResource>();

		//no volume loaded yet -> return
		if (!pVolume || pVolume->GetSize().GetMaximum() == 0)
			return;

		// Default bounding box
		Box boxBounds = pVolume->GetBounds().GetTranslated(-pVolume->GetBounds().GetCenter());
		boxBounds = pVolume->GetProperty("Bounding Box",boxBounds);

		const bool bUseBoundingBox = GetPlugin().GetProperty("Use Bounding Box");

		if (bUseBoundingBox)
			boxBounds = pVolume->GetProperty("Bounding Box",boxBounds);

		//Calculate bounding box in world coordinates (minimum point = start, extent = range)
		m_fStartX = boxBounds.GetMinimum().GetX();
		m_fStartY = boxBounds.GetMinimum().GetY();
		m_fStartZ = boxBounds.GetMinimum().GetZ();
		m_fRangeX = boxBounds.GetMaximum().GetX() - m_fStartX;
		m_fRangeY = boxBounds.GetMaximum().GetY() - m_fStartY;
		m_fRangeZ = boxBounds.GetMaximum().GetZ() - m_fStartZ;

		m_glyphPositions.clear();
		m_uElementCount = 0;


		m_fGlyphDensity = GetPlugin().GetProperty("Glyph Density");

		Vector sizes = pVolume->GetSize();
		int iNumSamplesX = int(sizes[0]);
		int iNumSamplesY = int(sizes[1]);
		int iNumSamplesZ = int(sizes[2]);

		//Derive number of glyphs to be displayed in each dimension x,y and z
		m_iNumGlyphsX = std::max(1, int( float(iNumSamplesX) * m_fGlyphDensity));
		m_iNumGlyphsY = std::max(1, int( float(iNumSamplesY) * m_fGlyphDensity));
		m_iNumGlyphsZ = std::max(1, int( float(iNumSamplesZ) * m_fGlyphDensity));

		//Now calculate and store world positions of glyphs in regular grid
		for(int x=0; x<m_iNumGlyphsX; x++)
		{	
			for(int y=0; y<m_iNumGlyphsY; y++)
			{	
				for(int z=0; z<m_iNumGlyphsZ; z++)
				{

					m_glyphPositions.push_back(m_fStartX + m_fRangeX * float(x)/float(m_iNumGlyphsX));
					m_glyphPositions.push_back(m_fStartY + m_fRangeY * float(y)/float(m_iNumGlyphsY));
					m_glyphPositions.push_back(m_fStartZ + m_fRangeZ * float(z)/float(m_iNumGlyphsZ));


					m_uElementCount++;
				}
			}
		}

		glBindBuffer(GL_ARRAY_BUFFER, m_uVertexBuffer);
		glBufferData(GL_ARRAY_BUFFER, m_glyphPositions.size()*sizeof(float), (void*)&(m_glyphPositions.front()), GL_STATIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, 0);

		//The glyph (point) size depends on the number of glyphs to be displayed.
		//It is derived from the smallest volume dimension (x,y or z) divided by the largest number of glyphs (x,y or z) to make sure, that glyphs don't overlap
		m_fPointSize = std::min(m_fRangeX, std::min(m_fRangeY, m_fRangeZ)) / float(std::max(m_iNumGlyphsX, std::max(m_iNumGlyphsY, m_iNumGlyphsZ)));

		m_bUpdateVertexBuffer = false;
	}
}


/**
 * @copydoc Renderer::button()
 */
bool RendererVisLabGlyphExercise::button(const Controller & conController)
{
	return false;
}

/**
 * @copydoc Renderer::cursor()
 */
bool RendererVisLabGlyphExercise::cursor(const Controller & conController)
{
	return false;
}

/**
 * @copydoc Renderer::GetButtonRole()
 */
const std::string RendererVisLabGlyphExercise::GetButtonRole(const unsigned int uIndex) const
{
	return std::string();
}

/**
 * @copydoc Renderer::GetCursorRole()
 */
const std::string RendererVisLabGlyphExercise::GetCursorRole(const unsigned int uIndex) const
{
	return std::string();
}
